""" Generic Heat Source (Oil, Gas or DistrictHeating) together with Configuration and State. """

# clean

# Import packages from standard library or the environment e.g. pandas, numpy etc.
import importlib
from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import dataclass_json
import pandas as pd

# Import modules from HiSim
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import log
from hisim.components import controller_l1_heatpump
from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry, KpiTagEnumClass
from hisim.components.configuration import HouseholdWarmWaterDemandConfig, EmissionFactorsAndCostsForFuelsConfig
from hisim.component import OpexCostDataClass, CapexCostDataClass

__authors__ = "Johanna Ganglbauer - johanna.ganglbauer@4wardenergy.at"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class HeatSourceConfig(cp.ConfigBase):
    """Configuration of a generic HeatSource."""

    building_name: str
    #: name of the device
    name: str
    #: priority of the device in hierachy: the higher the number the lower the priority
    source_weight: int
    #: type of heat source classified by fuel (Oil, Gas or DistrictHeating)
    fuel: lt.LoadTypes
    #: maximal thermal power of heat source in kW
    power_th: float
    #: usage of the heatpump: either for heating or for water heating
    water_vs_heating: lt.InandOutputType
    #: efficiency of the fuel to heat conversion
    efficiency: float
    #: CO2 footprint of investment in kg
    co2_footprint: float
    #: cost for investment in Euro
    cost: float
    #: lifetime in years
    lifetime: float

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return HeatSource.get_full_classname()

    @classmethod
    def get_default_config_heating(
        cls,
        building_name: str = "BUI1",
    ) -> "HeatSourceConfig":
        """Returns default configuration of a Heat Source used for heating."""
        config = HeatSourceConfig(
            building_name=building_name,
            name="HeatingHeatSource",
            source_weight=1,
            fuel=lt.LoadTypes.DISTRICTHEATING,
            power_th=6200.0,
            water_vs_heating=lt.InandOutputType.HEATING,
            efficiency=1.0,
            co2_footprint=0,
            cost=0,
            lifetime=1
        )
        return config

    @classmethod
    def get_default_config_waterheating_with_district_heating(
        cls,
        building_name: str = "BUI1",
    ) -> "HeatSourceConfig":
        """Returns default configuration of a Heat Source used for water heating (DHW)."""
        config = HeatSourceConfig(
            building_name=building_name,
            name="DHWHeatSource",
            source_weight=1,
            fuel=lt.LoadTypes.DISTRICTHEATING,
            power_th=3000.0,
            water_vs_heating=lt.InandOutputType.WATER_HEATING,
            efficiency=1.0,
            co2_footprint=0,
            cost=0,
            lifetime=1
        )
        return config

    @classmethod
    def get_default_config_waterheating_with_gas(
        cls,
        max_warm_water_demand_in_liter: float,
        scaling_factor_according_to_number_of_apartments: float,
        seconds_per_timestep: int,
        building_name: str = "BUI1",
    ) -> "HeatSourceConfig":
        """Returns default configuration of a Heat Source used for water heating (DHW)."""
        # use importlib for importing the other component in order to avoid circular-import errors

        component_module_name = "hisim.modular_household.component_connections"
        component_module = importlib.import_module(name=component_module_name)
        get_heating_system_efficiency = getattr(component_module, "get_heating_system_efficiency")
        thermal_power_in_watt = (
            max_warm_water_demand_in_liter
            * (4180 / 3600)
            * 0.5
            * (3600 / seconds_per_timestep)
            * (
                HouseholdWarmWaterDemandConfig.ww_temperature_demand
                - HouseholdWarmWaterDemandConfig.temperature_difference_hot
                - HouseholdWarmWaterDemandConfig.freshwater_temperature
            )
            * scaling_factor_according_to_number_of_apartments
        )
        efficiency = get_heating_system_efficiency(
            heating_system_installed=lt.HeatingSystems.GAS_HEATING,
            water_vs_heating=lt.InandOutputType.HEATING,
        )
        config = HeatSourceConfig(
            building_name=building_name,
            name="DHWHeatSource",
            source_weight=1,
            fuel=lt.LoadTypes.GAS,
            power_th=thermal_power_in_watt,
            water_vs_heating=lt.InandOutputType.WATER_HEATING,
            efficiency=efficiency,
            co2_footprint=thermal_power_in_watt * 1e-3 * 49.47,  # value from emission_factros_and_costs_devices.csv
            cost=7416,  # value from emission_factros_and_costs_devices.csv
            lifetime=20,  # value from emission_factros_and_costs_devices.csv
        )
        return config


class HeatSourceState:
    """Heat source state class saves the state of the heat source."""

    def __init__(self, state: int = 0):
        """Initializes state."""
        self.state = state

    def clone(self) -> "HeatSourceState":
        """Creates copy of a state."""
        return HeatSourceState(state=self.state)


class HeatSource(cp.Component):
    """Heat Source implementation.

    District Heating, Oil Heating or Gas Heating. Heat is converted with given efficiency.

    Components to connect to:
    (1) Heat Pump Controller (controller_l1_heatpump)
    """

    # Inputs
    L1HeatSourceTargetPercentage = "L1HeatSourceTargetPercentage"

    # Outputs
    ThermalPowerDelivered = "ThermalPowerDelivered"
    FuelDelivered = "FuelDelivered"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: HeatSourceConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        # introduce parameters of district heating
        self.config = config
        self.state = HeatSourceState()
        self.previous_state = HeatSourceState()

        # Inputs - Mandatories
        self.l1_heatsource_taget_percentage: cp.ComponentInput = self.add_input(
            self.component_name,
            self.L1HeatSourceTargetPercentage,
            lt.LoadTypes.ANY,
            lt.Units.PERCENT,
            mandatory=True,
        )

        # Outputs
        self.thermal_power_delivered_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ThermalPowerDelivered,
            load_type=lt.LoadTypes.HEATING,
            unit=lt.Units.WATT,
            postprocessing_flag=[lt.InandOutputType.THERMAL_PRODUCTION],
            output_description="Thermal Power Delivered",
        )
        self.fuel_delivered_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.FuelDelivered,
            load_type=self.config.fuel,
            unit=lt.Units.ANY,
            postprocessing_flag=[
                lt.InandOutputType.FUEL_CONSUMPTION,
                config.fuel,
                config.water_vs_heating,
            ],
            output_description="Fuel Delivered",
        )

        if config.fuel == lt.LoadTypes.OIL:
            self.fuel_delivered_channel.unit = lt.Units.LITER
        else:
            self.fuel_delivered_channel.unit = lt.Units.WATT_HOUR

        self.add_default_connections(self.get_default_connections_controller_l1_heatpump())

    def get_default_connections_controller_l1_heatpump(
        self,
    ) -> List[cp.ComponentConnection]:
        """Sets default connections of heat source controller."""
        log.information("setting l1 default connections in Generic Heat Source")
        connections = []
        controller_classname = controller_l1_heatpump.L1HeatPumpController.get_classname()
        connections.append(
            cp.ComponentConnection(
                HeatSource.L1HeatSourceTargetPercentage,
                controller_classname,
                controller_l1_heatpump.L1HeatPumpController.HeatControllerTargetPercentage,
            )
        )
        return connections

    def write_to_report(self) -> List[str]:
        """Writes relevant data to report."""

        lines = []
        lines.append(f"Name: {self.config.name + str(self.config.source_weight)})")
        lines.append(f"Fuel: {self.config.fuel}")
        lines.append(f"Power: {(self.config.power_th) * 1e-3:4.0f} kW")
        lines.append(f"Efficiency : {(self.config.efficiency) * 100:4.0f} %")
        return lines

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Performs the simulation of the heat source model."""

        # Inputs
        target_percentage = stsv.get_input_value(self.l1_heatsource_taget_percentage)

        # calculate modulation
        if target_percentage > 0:
            power_modifier = target_percentage
        if target_percentage == 0:
            power_modifier = 0
        if target_percentage < 0:
            power_modifier = 0
        if power_modifier > 1:
            power_modifier = min(power_modifier, 1)

        stsv.set_output_value(
            self.thermal_power_delivered_channel,
            self.config.power_th * power_modifier * self.config.efficiency,
        )

        if self.config.fuel == lt.LoadTypes.OIL:
            # conversion from Wh oil to liter oil
            stsv.set_output_value(
                self.fuel_delivered_channel,
                power_modifier
                * self.config.power_th
                * 1.0526315789474e-4
                * self.my_simulation_parameters.seconds_per_timestep
                / 3.6e3,
            )
        else:
            stsv.set_output_value(
                self.fuel_delivered_channel,
                power_modifier * self.config.power_th * self.my_simulation_parameters.seconds_per_timestep / 3.6e3,
            )

    @staticmethod
    def get_cost_capex(config: HeatSourceConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:
        """Returns investment cost, CO2 emissions and lifetime."""
        if config.fuel == lt.LoadTypes.GAS:
            seconds_per_year = 365 * 24 * 60 * 60
            capex_per_simulated_period = (config.cost / config.lifetime) * (
                simulation_parameters.duration.total_seconds() / seconds_per_year
            )
            device_co2_footprint_per_simulated_period = (config.co2_footprint / config.lifetime) * (
                simulation_parameters.duration.total_seconds() / seconds_per_year
            )

            capex_cost_data_class = CapexCostDataClass(
                capex_investment_cost_in_euro=config.cost,
                device_co2_footprint_in_kg=config.co2_footprint,
                lifetime_in_years=config.lifetime,
                capex_investment_cost_for_simulated_period_in_euro=capex_per_simulated_period,
                device_co2_footprint_for_simulated_period_in_kg=device_co2_footprint_per_simulated_period,
                kpi_tag=KpiTagEnumClass.GAS_HEATER_DOMESTIC_HOT_WATER
            )
            return capex_cost_data_class
        return NotImplemented

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of energy and maintenance costs."""
        gas_consumption_in_kilowatt_hour: Optional[float] = None
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name and output.load_type == lt.LoadTypes.GAS:
                gas_consumption_in_kilowatt_hour = round(sum(postprocessing_results.iloc[:, index]) * 1e-3, 1)
        if gas_consumption_in_kilowatt_hour is not None:
            emissions_and_cost_factors = EmissionFactorsAndCostsForFuelsConfig.get_values_for_year(
                self.my_simulation_parameters.year
            )
            co2_per_unit = emissions_and_cost_factors.gas_footprint_in_kg_per_kwh
            co2_per_simulated_period_in_kg = gas_consumption_in_kilowatt_hour * co2_per_unit

            # energy costs and co2 and everything will be considered in gas meter
            opex_cost_data_class = OpexCostDataClass(
                opex_energy_cost_in_euro=0,
                opex_maintenance_cost_in_euro=0,  # TODO: needs o be implemented still
                co2_footprint_in_kg=co2_per_simulated_period_in_kg,
                consumption_in_kwh=gas_consumption_in_kilowatt_hour,
                loadtype=lt.LoadTypes.GAS,
                kpi_tag=KpiTagEnumClass.GAS_HEATER_DOMESTIC_HOT_WATER
            )
        else:
            opex_cost_data_class = OpexCostDataClass.get_default_opex_cost_data_class()

        return opex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        gas_consumption_in_kilowatt_hour: Optional[float] = None
        list_of_kpi_entries: List[KpiEntry] = []
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name:
                if output.field_name == self.FuelDelivered and output.load_type == lt.LoadTypes.GAS:
                    gas_consumption_in_kilowatt_hour = round(postprocessing_results.iloc[:, index].sum() * 1e-3, 1)
                    break
        my_kpi_entry = KpiEntry(
            name="Gas consumption for domestic hot water",
            unit="kWh",
            value=gas_consumption_in_kilowatt_hour,
            tag=KpiTagEnumClass.GAS_HEATER_DOMESTIC_HOT_WATER,
            description=self.component_name,
        )

        list_of_kpi_entries.append(my_kpi_entry)
        return list_of_kpi_entries



"""Calculate Opex and Capex for each component."""

import os
from typing import List

import pandas as pd
from hisim import log
from hisim.simulationparameters import SimulationParameters
from hisim.component_wrapper import ComponentWrapper
from hisim.component import OpexCostDataClass, CapexCostDataClass
from hisim.components.electricity_meter import ElectricityMeter
from hisim.components.gas_meter import GasMeter
from hisim.components.heating_meter import HeatingMeter
from hisim.components.more_advanced_heat_pump_hplib import MoreAdvancedHeatPumpHPLib
from hisim.components.advanced_heat_pump_hplib import HeatPumpHplib
from hisim.components.generic_heat_pump_modular import ModularHeatPump
from hisim.components.simple_const_heat_source import SimpleHeatSource


def opex_calculation(
    components: List[ComponentWrapper],
    all_outputs: List,
    postprocessing_results: pd.DataFrame,
    simulation_parameters: SimulationParameters,
    building_objects_in_district_list: list,
) -> List:
    """Loops over all components and calls opex cost calculation."""
    total_operational_co2_footprint = 0.0
    total_opex_energy_cost = 0.0
    total_opex_maintenance_cost = 0.0
    total_consumption_in_kwh = 0.0
    total_operational_co2_footprint_without_hp = 0.0
    total_opex_energy_cost_without_hp = 0.0
    total_opex_maintenance_cost_without_hp = 0.0
    total_consumption_in_kwh_without_hp = 0.0

    headline: List[object] = [
        "Component",
        "Maintenance Costs in EUR",
        "Fuel CO2 Emissions in kg",
        "Consumption in kWh",
    ]
    opex_table_as_list_of_list = []

    for building_object in building_objects_in_district_list:
        total_energy_cost_building_object = 0.0
        total_maintenance_cost_building_object = 0.0
        total_operational_co2_footprint_building_object = 0.0
        total_consumption_in_kwh_building_object = 0.0
        total_energy_cost_building_object_without_hp = 0.0
        total_maintenance_cost_building_object_without_hp = 0.0
        total_operational_co2_footprint_building_object_without_hp = 0.0
        total_consumption_in_kwh_building_object_without_hp = 0.0
        for component in components:
            component_unwrapped = component.my_component
            if (building_object in str(component_unwrapped.component_name) or
                    not simulation_parameters.multiple_buildings):
                # cost and co2_footprint are calculated per simulated period
                component_class_name = component_unwrapped.get_classname()
                # get opex data for all components except Electricity and Gas Meter
                if (
                    component_class_name != ElectricityMeter.get_classname()
                    and component_class_name != GasMeter.get_classname()
                    and component_class_name != HeatingMeter.get_classname()
                ):
                    opex_cost_data_class: OpexCostDataClass = component_unwrapped.get_cost_opex(
                        all_outputs=all_outputs,
                        postprocessing_results=postprocessing_results,
                    )
                    cost_energy = opex_cost_data_class.opex_energy_cost_in_euro
                    cost_maintenance = opex_cost_data_class.opex_maintenance_cost_in_euro
                    co2_footprint = opex_cost_data_class.co2_footprint_in_kg
                    consumption = opex_cost_data_class.consumption_in_kwh
                    total_energy_cost_building_object += cost_energy
                    total_maintenance_cost_building_object += cost_maintenance
                    total_operational_co2_footprint_building_object += co2_footprint
                    total_consumption_in_kwh_building_object += consumption

                    if isinstance(
                        component_unwrapped,
                        (HeatPumpHplib, ModularHeatPump, MoreAdvancedHeatPumpHPLib, SimpleHeatSource),
                    ):
                        pass
                    else:
                        total_energy_cost_building_object_without_hp += cost_energy
                        total_maintenance_cost_building_object_without_hp += cost_maintenance
                        total_operational_co2_footprint_building_object_without_hp += co2_footprint
                        total_consumption_in_kwh_building_object_without_hp += consumption

                    opex_table_as_list_of_list.append(
                        [
                            component_unwrapped.component_name,
                            round(cost_maintenance, 2),
                            round(co2_footprint, 2),
                            round(consumption, 2),
                        ]
                    )
        if simulation_parameters.multiple_buildings:
            opex_table_as_list_of_list.append(
                [
                    f"{building_object}_Total",
                    round(total_maintenance_cost_building_object, 2),
                    round(total_operational_co2_footprint_building_object, 2),
                    round(total_consumption_in_kwh_building_object, 2),
                ]
            )

            opex_table_as_list_of_list.append(
                [
                    f"{building_object}_Total_without_heatpump",
                    round(total_maintenance_cost_building_object_without_hp, 2),
                    round(total_operational_co2_footprint_building_object_without_hp, 2),
                    round(total_consumption_in_kwh_building_object_without_hp, 2),
                ]
            )

            opex_table_as_list_of_list.append(
                [
                    f"{building_object}_Total_only_heatpump",
                    round(total_maintenance_cost_building_object - total_maintenance_cost_building_object_without_hp, 2),
                    round(
                        total_operational_co2_footprint_building_object
                        - total_operational_co2_footprint_building_object_without_hp,
                        2,
                    ),
                    round(
                        total_consumption_in_kwh_building_object - total_consumption_in_kwh_building_object_without_hp, 2
                    ),
                ]
            )

        total_opex_energy_cost += total_energy_cost_building_object
        total_opex_maintenance_cost += total_maintenance_cost_building_object
        total_operational_co2_footprint += total_operational_co2_footprint_building_object
        total_consumption_in_kwh += total_consumption_in_kwh_building_object

        total_opex_energy_cost_without_hp += total_energy_cost_building_object_without_hp
        total_opex_maintenance_cost_without_hp += total_maintenance_cost_building_object_without_hp
        total_operational_co2_footprint_without_hp += total_operational_co2_footprint_building_object_without_hp
        total_consumption_in_kwh_without_hp += total_consumption_in_kwh_building_object_without_hp

    opex_table_as_list_of_list.append(
        [
            "Total",
            round(total_opex_maintenance_cost, 2),
            round(total_operational_co2_footprint, 2),
            round(total_consumption_in_kwh, 2),
        ]
    )
    opex_table_as_list_of_list.append(
        [
            "Total_without_heatpump",
            round(total_opex_maintenance_cost_without_hp, 2),
            round(total_operational_co2_footprint_without_hp, 2),
            round(total_consumption_in_kwh_without_hp, 2),
        ]
    )
    opex_table_as_list_of_list.append(
        [
            "Total_only_heatpump",
            round(total_opex_maintenance_cost - total_opex_maintenance_cost_without_hp, 2),
            round(total_operational_co2_footprint - total_operational_co2_footprint_without_hp, 2),
            round(total_consumption_in_kwh - total_consumption_in_kwh_without_hp, 2),
        ]
    )

    pathname = os.path.join(simulation_parameters.result_directory, "operational_costs_co2_footprint.csv")
    opex_df = pd.DataFrame(opex_table_as_list_of_list, columns=headline)
    opex_df.to_csv(pathname, index=False, header=True, encoding="utf8", sep=";")

    opex_table_as_list_of_list.insert(0, headline)

    return opex_table_as_list_of_list


def capex_calculation(
    components: List[ComponentWrapper],
    simulation_parameters: SimulationParameters,
    building_objects_in_district_list: list,
) -> List:
    """Loops over all components and calls capex cost calculation."""
    total_investment_cost = 0.0
    total_device_co2_footprint = 0.0
    total_investment_cost_per_simulated_period = 0.0
    total_device_co2_footprint_per_simulated_period = 0.0
    total_investment_cost_without_hp = 0.0
    total_device_co2_footprint_without_hp = 0.0
    total_investment_cost_per_simulated_period_without_hp = 0.0
    total_device_co2_footprint_per_simulated_period_without_hp = 0.0

    headline: List[object] = [
        "Component",
        "Investment in EUR",
        "Device CO2-footprint in kg",
        "Lifetime in years",
    ]
    capex_table_as_list_of_list = []
    for building_object in building_objects_in_district_list:
        total_investment_cost_building_object = 0.0
        total_device_co2_footprint_building_object = 0.0
        total_investment_cost_per_simulated_period_building_object = 0.0
        total_device_co2_footprint_per_simulated_period_building_object = 0.0
        total_investment_cost_building_object_without_hp = 0.0
        total_device_co2_footprint_building_object_without_hp = 0.0
        total_investment_cost_per_simulated_period_building_object_without_hp = 0.0
        total_device_co2_footprint_per_simulated_period_building_object_without_hp = 0.0
        for component in components:
            component_unwrapped = component.my_component
            # capex, co2_footprint, lifetime = component_unwrapped.get_cost_capex(
            #     config=component_unwrapped.config,
            # )
            if (building_object in str(component_unwrapped.component_name) or
                    not simulation_parameters.multiple_buildings):
                capex_cost_data_class: CapexCostDataClass = component_unwrapped.get_cost_capex(
                    config=component_unwrapped.config, simulation_parameters=simulation_parameters
                )

                if capex_cost_data_class.lifetime_in_years > 0:

                    total_investment_cost_building_object += capex_cost_data_class.capex_investment_cost_in_euro
                    total_device_co2_footprint_building_object += capex_cost_data_class.device_co2_footprint_in_kg
                    total_investment_cost_per_simulated_period_building_object += capex_cost_data_class.capex_investment_cost_for_simulated_period_in_euro
                    total_device_co2_footprint_per_simulated_period_building_object += capex_cost_data_class.device_co2_footprint_for_simulated_period_in_kg

                    if isinstance(
                        component_unwrapped,
                        (HeatPumpHplib, ModularHeatPump, MoreAdvancedHeatPumpHPLib, SimpleHeatSource),
                    ):
                        pass
                    else:
                        total_investment_cost_building_object_without_hp += capex_cost_data_class.capex_investment_cost_in_euro
                        total_device_co2_footprint_building_object_without_hp += capex_cost_data_class.device_co2_footprint_in_kg
                        total_investment_cost_per_simulated_period_building_object_without_hp += (
                            capex_cost_data_class.capex_investment_cost_for_simulated_period_in_euro
                        )
                        total_device_co2_footprint_per_simulated_period_building_object_without_hp += (
                            capex_cost_data_class.device_co2_footprint_for_simulated_period_in_kg
                        )

                    capex_table_as_list_of_list.append(
                        [
                            component_unwrapped.component_name,
                            round(capex_cost_data_class.capex_investment_cost_for_simulated_period_in_euro, 2),
                            round(capex_cost_data_class.device_co2_footprint_for_simulated_period_in_kg, 2),
                            capex_cost_data_class.lifetime_in_years,
                        ]
                    )
                else:
                    log.warning(f"capex calculation not valid. Check lifetime in Configuration of {component}")

        if simulation_parameters.multiple_buildings:
            capex_table_as_list_of_list.append(
                [
                    f"{building_object}_Total_per_simulated_period_without_heatpump",
                    round(total_investment_cost_per_simulated_period_building_object_without_hp, 2),
                    round(total_device_co2_footprint_per_simulated_period_building_object_without_hp, 2),
                    "---",
                ]
            )
            capex_table_as_list_of_list.append(
                [
                    f"{building_object}_Total_per_simulated_period_only_heatpump",
                    round(
                        total_investment_cost_per_simulated_period_building_object
                        - total_investment_cost_per_simulated_period_building_object_without_hp,
                        2,
                    ),
                    round(
                        total_device_co2_footprint_per_simulated_period_building_object
                        - total_device_co2_footprint_per_simulated_period_building_object_without_hp,
                        2,
                    ),
                    "---",
                ]
            )
            capex_table_as_list_of_list.append(
                [
                    f"{building_object}_Total_per_simulated_period",
                    round(total_investment_cost_per_simulated_period_building_object, 2),
                    round(total_device_co2_footprint_per_simulated_period_building_object, 2),
                    "---",
                ]
            )
            capex_table_as_list_of_list.append(
                [
                    f"{building_object}_Total_without_heatpump",
                    round(total_investment_cost_building_object_without_hp, 2),
                    round(total_device_co2_footprint_building_object_without_hp, 2),
                    "---",
                ]
            )
            capex_table_as_list_of_list.append(
                [
                    f"{building_object}_Total_only_heatpump",
                    round(total_investment_cost_building_object - total_investment_cost_building_object_without_hp, 2),
                    round(
                        total_device_co2_footprint_building_object - total_device_co2_footprint_building_object_without_hp,
                        2,
                    ),
                    "---",
                ]
            )
            capex_table_as_list_of_list.append(
                [
                    f"{building_object}_Total",
                    round(total_investment_cost_building_object, 2),
                    round(total_device_co2_footprint_building_object, 2),
                    "---",
                ]
            )

        total_investment_cost += total_investment_cost_building_object
        total_device_co2_footprint += total_device_co2_footprint_building_object
        total_investment_cost_per_simulated_period += total_investment_cost_per_simulated_period_building_object
        total_device_co2_footprint_per_simulated_period += (
            total_device_co2_footprint_per_simulated_period_building_object
        )

        total_investment_cost_without_hp += total_investment_cost_building_object_without_hp
        total_device_co2_footprint_without_hp += total_device_co2_footprint_building_object_without_hp
        total_investment_cost_per_simulated_period_without_hp += (
            total_investment_cost_per_simulated_period_building_object_without_hp
        )
        total_device_co2_footprint_per_simulated_period_without_hp += (
            total_device_co2_footprint_per_simulated_period_building_object_without_hp
        )

    capex_table_as_list_of_list.append(
        [
            "Total_per_simulated_period",
            round(total_investment_cost_per_simulated_period, 2),
            round(total_device_co2_footprint_per_simulated_period, 2),
            "---",
        ]
    )
    capex_table_as_list_of_list.append(
        [
            "Total_per_simulated_period_without_heatpump",
            round(total_investment_cost_per_simulated_period_without_hp, 2),
            round(total_device_co2_footprint_per_simulated_period_without_hp, 2),
            "---",
        ]
    )
    capex_table_as_list_of_list.append(
        [
            "Total_per_simulated_period_only_heatpump",
            round(
                total_investment_cost_per_simulated_period - total_investment_cost_per_simulated_period_without_hp, 2
            ),
            round(
                total_device_co2_footprint_per_simulated_period
                - total_device_co2_footprint_per_simulated_period_without_hp,
                2,
            ),
            "---",
        ]
    )
    capex_table_as_list_of_list.append(
        [
            "Total",
            round(total_investment_cost, 2),
            round(total_device_co2_footprint, 2),
            "---",
        ]
    )
    capex_table_as_list_of_list.append(
        [
            "Total_without_heatpump",
            round(total_investment_cost_without_hp, 2),
            round(total_device_co2_footprint_without_hp, 2),
            "---",
        ]
    )
    capex_table_as_list_of_list.append(
        [
            "Total_only_heatpump",
            round(total_investment_cost - total_investment_cost_without_hp, 2),
            round(total_device_co2_footprint - total_device_co2_footprint_without_hp, 2),
            "---",
        ]
    )

    pathname = os.path.join(simulation_parameters.result_directory, "investment_cost_co2_footprint.csv")
    capex_df = pd.DataFrame(capex_table_as_list_of_list, columns=headline)
    capex_df.to_csv(pathname, index=False, header=True, encoding="utf8", sep=";")

    capex_table_as_list_of_list.insert(0, headline)

    return capex_table_as_list_of_list



"""Controller l1 example module."""

# clean

# Generic/Built-in

from typing import Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json

# Owned
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput, ConfigBase, DisplayConfig
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters

__authors__ = "Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class SimpleControllerConfig(ConfigBase):
    """Config class."""

    building_name: str
    name: str

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return SimpleController.get_full_classname()

    @classmethod
    def get_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Returns default config."""
        config = SimpleControllerConfig(name="SimpleController", building_name=building_name)
        return config


class SimpleController(Component):
    """Simple controller class."""

    StorageFillLevel = "Fill Level Percent"
    GasHeaterPowerPercent = "Gas Heater Power Level"

    def __init__(
        self,
        name: str,
        my_simulation_parameters: SimulationParameters,
        config: SimpleControllerConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ) -> None:
        """Initialize the class."""

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.input1_channel: ComponentInput = self.add_input(
            self.component_name,
            SimpleController.StorageFillLevel,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KWH,
            True,
        )
        self.output1_channel: ComponentOutput = self.add_output(
            self.component_name,
            SimpleController.GasHeaterPowerPercent,
            lt.LoadTypes.GAS,
            lt.Units.PERCENT,
        )
        self.state = 0
        self.previous_state = self.state

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""

        if force_convergence:
            return
        percent = stsv.get_input_value(self.input1_channel)
        if percent < 0.4:
            self.state = 1
        if percent > 0.99:
            self.state = 0
        stsv.set_output_value(self.output1_channel, self.state)



"""Generic Heat Pump module.

This module contains the following classes:
    1. GenericHeatPump State
    2. GenericHeatPump
    3. HeatPumpController

"""

# clean

# Generic/Built-in
import copy
from typing import List, Any, Optional
from dataclasses import dataclass
import numpy as np
from dataclasses_json import dataclass_json
import pandas as pd

from hisim import component as cp
from hisim import log

# Owned
from hisim import utils
from hisim.component import OpexCostDataClass, CapexCostDataClass
from hisim.components.building import Building
from hisim.components.weather import Weather
import hisim.loadtypes as lt
from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry

__authors__ = "Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class GenericHeatPumpConfig(cp.ConfigBase):
    """Config for the generic heat pump."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return GenericHeatPump.get_full_classname()

    building_name: str
    name: str
    manufacturer: str
    heat_pump_name: str
    min_operation_time: float
    min_idle_time: float

    @classmethod
    def get_default_generic_heat_pump_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default Generic Heat Pump."""
        return GenericHeatPumpConfig(
            building_name=building_name,
            name="HeatPump",
            heat_pump_name="Vitocal 300-A AWO-AC 301.B07",
            manufacturer="Viessmann Werke GmbH & Co KG",
            min_operation_time=60 * 60,
            min_idle_time=15 * 60,
        )


@dataclass_json
@dataclass
class GenericHeatPumpControllerConfig(cp.ConfigBase):
    """Controller for the generic heat pump."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return GenericHeatPumpController.get_full_classname()

    building_name: str
    name: str
    temperature_air_heating_in_celsius: float
    temperature_air_cooling_in_celsius: float
    offset: float
    mode: int

    @classmethod
    def get_default_generic_heat_pump_controller_config(cls, building_name: str = "BUI1",) -> Any:
        """Gets a default Generic Heat Pump Controller."""
        return GenericHeatPumpControllerConfig(
            building_name=building_name,
            name="HeatPumpController",
            temperature_air_heating_in_celsius=18.0,
            temperature_air_cooling_in_celsius=26.0,
            offset=0.5,
            mode=1,
        )


class GenericHeatPumpState:
    """Heat Pump State class.

    It determines the state of the heat pump.

    """

    def __init__(
        self,
        start_timestep: int = 0,
        thermal_power_delivered_in_watt: float = 0.0,
        cop: float = 1.0,
        cycle_number: Optional[int] = None,
    ) -> None:
        """Contruct all the necessary attributes."""
        self.start_timestep = start_timestep
        self.thermal_power_delivered_in_watt = thermal_power_delivered_in_watt
        self.cycle_number = cycle_number
        if thermal_power_delivered_in_watt == 0.0:
            self.activation = 0
            self.heating_power_in_watt = 0.0
            self.cooling_power_in_watt = 0.0
            self.cop = 1.0
            self.electricity_input_in_watt = abs(self.thermal_power_delivered_in_watt / self.cop)
        elif self.thermal_power_delivered_in_watt > 0.0:
            self.activation = -1
            self.heating_power_in_watt = self.thermal_power_delivered_in_watt
            self.cooling_power_in_watt = 0.0
            self.cop = cop
            self.electricity_input_in_watt = abs(self.thermal_power_delivered_in_watt / self.cop)
        elif self.thermal_power_delivered_in_watt < 0.0:
            self.activation = 1
            self.heating_power_in_watt = 0
            self.cooling_power_in_watt = self.thermal_power_delivered_in_watt
            self.cop = cop
            self.electricity_input_in_watt = abs(self.thermal_power_delivered_in_watt / self.cop)
        else:
            raise Exception("Impossible Heat Pump State.")

    def clone(self) -> Any:
        """Clone heat pump state."""
        return GenericHeatPumpState(
            self.start_timestep,
            self.thermal_power_delivered_in_watt,
            self.cop,
            self.cycle_number,
        )


class GenericHeatPump(cp.Component):
    """Heat pump class.

    It does support a refrigeration cycle.
    Thermal output is delivered straight to
    the component object.

    Parameters
    ----------
    manufacturer : str
        Heat pump manufacturer
    name : str
        Heat pump model
    min_operation_time : int, optional
        Minimum time duration that the heat pump operates under one cycle, in seconds. The default is 3600.
    min_idle_time : int, optional
        Minimum time duration that the heat pump has to stay idle, in seconds. The default is 900.

    """

    # Inputs
    State = "State"
    TemperatureOutside = "TemperatureOutside"
    WaterConsumption = "WaterConsumption"
    WaterInput_mass = "WaterInput_mass"  # kg/s
    WaterInput_temperature = "WaterInput_temperature"  # °C

    # Outputs
    ThermalPowerDelivered = "ThermalPowerDelivered"
    Heating = "Heating"
    Cooling = "Cooling"
    ElectricityOutput = "ElectricityOutput"
    NumberOfCycles = "NumberOfCycles"

    # Similar components to connect to:
    # 1. Weather
    # 2. HeatPumpController
    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: GenericHeatPumpConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Construct all the necessary attributes."""
        self.heatpump_config = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.manufacturer = self.heatpump_config.manufacturer
        self.heatpump_name = self.heatpump_config.heat_pump_name
        self.min_operation_time = self.heatpump_config.min_operation_time
        self.min_idle_time = self.heatpump_config.min_idle_time
        self.build(
            self.manufacturer,
            self.heatpump_name,
            self.min_operation_time,
            self.min_idle_time,
        )

        self.number_of_cycles = 0
        self.number_of_cycles_previous = copy.deepcopy(self.number_of_cycles)
        self.state = GenericHeatPumpState(start_timestep=int(0), cycle_number=0)
        self.previous_state = self.state.clone()
        self.has_been_converted: Any
        # Inputs - Mandatories
        self.state_channel: cp.ComponentInput = self.add_input(
            self.component_name, self.State, lt.LoadTypes.ANY, lt.Units.ANY, True
        )
        self.temperature_outside_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureOutside,
            lt.LoadTypes.ANY,
            lt.Units.CELSIUS,
            True,
        )
        # Inputs - Not Mandatories
        self.water_load_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.WaterConsumption,
            lt.LoadTypes.VOLUME,
            lt.Units.LITER,
            False,
        )
        self.water_input_mass_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.WaterInput_mass,
            lt.LoadTypes.WARM_WATER,
            lt.Units.KG_PER_SEC,
            False,
        )
        self.water_input_temperature_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.WaterInput_temperature,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            False,
        )

        # Outputs
        self.thermal_power_delivered_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.ThermalPowerDelivered,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            output_description=f"here a description for {self.ThermalPowerDelivered} will follow.",
        )

        self.heating_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.Heating,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            output_description=f"here a description for {self.Heating} will follow.",
        )

        self.cooling_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.Cooling,
            lt.LoadTypes.COOLING,
            lt.Units.WATT,
            output_description=f"here a description for {self.Cooling} will follow.",
        )

        self.electricity_output_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.ElectricityOutput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            output_description=f"here a description for Heat Pump {self.ElectricityOutput} will follow.",
            postprocessing_flag=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED],
        )

        self.number_of_cycles_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.NumberOfCycles,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description=f"here a description for {self.NumberOfCycles} will follow.",
        )

        self.add_default_connections(self.get_default_connections_from_weather())
        self.add_default_connections(self.get_default_connections_heatpump_controller())

    def get_default_connections_from_weather(self) -> List[cp.ComponentConnection]:
        """Get weather default connections."""

        connections = []
        weather_classname = Weather.get_classname()
        connections.append(
            cp.ComponentConnection(
                GenericHeatPump.TemperatureOutside,
                weather_classname,
                Weather.TemperatureOutside,
            )
        )
        return connections

    def get_default_connections_heatpump_controller(
        self,
    ) -> List[cp.ComponentConnection]:
        """Get heat pump controller default connections."""

        connections = []
        controller_classname = GenericHeatPumpController.get_classname()
        connections.append(
            cp.ComponentConnection(
                GenericHeatPump.State,
                controller_classname,
                GenericHeatPumpController.State,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepare the simulation."""
        pass

    def build(
        self,
        manufacturer: str,
        name: str,
        min_operation_time: float,
        min_idle_time: float,
    ) -> None:
        """Build function.

        The function sets important constants and parameters for the calculations.
        """
        # Simulation parameters

        # Retrieves heat pump from database - BEGIN
        heat_pumps_database = utils.load_smart_appliance("Heat Pump")

        heat_pump_found = False
        heat_pump = None
        for heat_pump in heat_pumps_database:
            if heat_pump["Manufacturer"] == manufacturer and heat_pump["Name"] == name:
                heat_pump_found = True
                break

        if not heat_pump_found or heat_pump is None:
            raise Exception("Heat pump model not registered in the database")

        # Interpolates COP data from the database
        self.cop_ref = []
        self.temperature_outside_ref = []
        for heat_pump_cops in heat_pump["COP"]:
            self.temperature_outside_ref.append(float([*heat_pump_cops][0][1:].split("/")[0]))
            self.cop_ref.append(float([*heat_pump_cops.values()][0]))
        self.cop_coef = np.polyfit(self.temperature_outside_ref, self.cop_ref, 1)

        self.max_heating_power_in_watt = heat_pump["Nominal Heating Power A2/35"] * 1e3
        # self.max_heating_power = 11 * 1E3
        self.max_cooling_power_in_watt = -self.max_heating_power_in_watt
        # Retrieves heat pump from database - END

        # Sets the power variation restrictions
        # Default values: 15 minutes to full power
        # Used only for non-clocked heat pump
        self.max_heating_power_variation_restriction_in_watt = (
            self.max_heating_power_in_watt * self.my_simulation_parameters.seconds_per_timestep / 900
        )
        self.max_cooling_power_variation_restriction_in_watt = (
            -self.max_heating_power_in_watt * self.my_simulation_parameters.seconds_per_timestep / 900
        )

        # Sets the time operation restricitions
        self.min_operation_time = min_operation_time / self.my_simulation_parameters.seconds_per_timestep
        self.min_idle_time = min_idle_time / self.my_simulation_parameters.seconds_per_timestep

        # Writes info to report
        self.write_to_report()

        # Applies correction due to timestep
        self.set_time_correction()  # self.set_time_correction(self.time_correction_factor)

    def set_time_correction(self, factor: float = 1) -> None:
        """Set time correction."""
        if factor == 1:
            self.has_been_converted = False
        if self.has_been_converted is True:
            raise Exception("It has been already converted!")
        self.max_heating_power_in_watt *= factor
        self.max_cooling_power_in_watt *= factor
        self.max_heating_power_variation_restriction_in_watt *= factor
        self.max_cooling_power_variation_restriction_in_watt *= factor
        if factor != 1:
            self.has_been_converted = True

    def calc_cop(self, t_out: float) -> float:
        """Calculate cop."""
        val: float = self.cop_coef[0] * t_out + self.cop_coef[1]
        return val

    def i_save_state(self) -> None:
        """Save the current state."""
        self.previous_state = self.state.clone()
        self.number_of_cycles_previous = self.number_of_cycles

    def i_restore_state(self) -> None:
        """Restore the previous state."""
        self.state = self.previous_state.clone()  # copy.deepcopy(self.previous_state)
        self.number_of_cycles = self.number_of_cycles_previous

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublecheck."""
        pass

    def write_to_report(self) -> List[str]:
        """Write important variables to report."""
        lines = []
        lines.append(f"Max Heating Power [kW]: {(self.max_heating_power_in_watt) * 1e-3:4.3f}")
        lines.append(
            f"Max Peating Power Variation Restriction [W]: {self.max_heating_power_variation_restriction_in_watt:4.3f}"
        )
        return self.heatpump_config.get_string_dict() + lines

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate heat pump."""
        # Inputs
        state_c = stsv.get_input_value(self.state_channel)
        temperature_outside = stsv.get_input_value(self.temperature_outside_channel)
        # log.information("State: {}, Temperature: {}".format(stateC, t_out))
        # log.information("State of Activation: {}".format(self.state.activation))
        # log.information("Timestep special: {}".format(self.state.start_timestep + self.min_idle_time))
        # Calculation

        # Calculation.ThermalEnergyStorage
        # ToDo: Implementation with Thermal Energy Storage - BEGIN
        # if self.water_loadC.SourceOutput is not None:
        #    if stsv.get_input_value(self.water_loadC) != 0:
        #        control_signal = 1
        #    else:
        #        control_signal = 0
        #    # Inputs
        #    water_input_mass_sec = stsv.get_input_value(self.water_input_mass)
        #    water_input_mass = water_input_mass_sec
        #    water_input_temperature = stsv.get_input_value(self.water_input_temperature)

        #    mass_flow_max = self.max_heating_power / (4180 * 25)  # kg/s ## -> ~0.07

        #    if control_signal == 1 and (water_input_mass == 0 and water_input_temperature == 0):
        #        """first iteration"""
        #        water_input_temperature = 40
        #        water_input_mass = mass_flow_max

        #    if control_signal == 1:
        #        volume_flow_gasheater = water_input_mass / PhysicsConfig.water_density
        #        ws = WaterSlice(WarmWaterStorageConfig.tank_diameter,
        #                       (4 * volume_flow_gasheater) / (pi * WarmWaterStorageConfig.tank_diameter ** 2),
        #                        water_input_temperature
        #                       )
        #        ws_output, wasted_energy_max_temperature, thermal_output = self.process_thermal(ws)
        #    else:
        #        height_flow_gasheater = 0
        #        water_input_temperature = 0
        #        ws = WaterSlice(WarmWaterStorageConfig.tank_diameter, height_flow_gasheater, water_input_temperature)
        #        ws_output = ws
        #        wasted_energy_max_temperature = 0

        #    ws_output_mass = ws_output.mass
        #    ws_output_temperature = ws_output.temperature

        #    # Mass is consistent
        #    stsv.set_output_value(self.water_output_mass, ws_output_mass)
        #    stsv.set_output_value(self.water_output_temperature, ws_output_temperature)
        #    stsv.set_output_value(self.wasted_energy_max_temperature, wasted_energy_max_temperature)
        # ToDo: Implementation with Thermal Energy Storage - END

        # Calculation.ThermalEnergyDelivery
        # Heat Pump is on
        if self.state.activation != 0:
            number_of_cycles = self.state.cycle_number
            # Checks if the minimum running time has been reached
            if timestep >= self.state.start_timestep + self.min_operation_time and state_c == 0:
                self.state = GenericHeatPumpState(start_timestep=timestep, cycle_number=number_of_cycles)
            stsv.set_output_value(
                self.thermal_power_delivered_channel,
                self.state.thermal_power_delivered_in_watt,
            )
            stsv.set_output_value(self.heating_channel, self.state.heating_power_in_watt)
            stsv.set_output_value(self.cooling_channel, self.state.cooling_power_in_watt)
            stsv.set_output_value(self.electricity_output_channel, self.state.electricity_input_in_watt)
            stsv.set_output_value(self.number_of_cycles_channel, self.number_of_cycles)
            return

        # Heat Pump is Off
        if state_c != 0 and (timestep >= self.state.start_timestep + self.min_idle_time):
            self.number_of_cycles = self.number_of_cycles + 1
            number_of_cycles = self.number_of_cycles
            if state_c == 1:
                # if stsv.get_input_value(self.stateC) > 0:
                self.state = GenericHeatPumpState(
                    start_timestep=timestep,
                    thermal_power_delivered_in_watt=self.max_heating_power_in_watt,
                    cop=self.calc_cop(temperature_outside),
                    cycle_number=number_of_cycles,
                )

            else:
                self.state = GenericHeatPumpState(
                    start_timestep=timestep,
                    thermal_power_delivered_in_watt=self.max_cooling_power_in_watt,
                    cop=self.calc_cop(temperature_outside),
                    cycle_number=number_of_cycles,
                )

        # Outputs
        stsv.set_output_value(
            self.thermal_power_delivered_channel,
            self.state.thermal_power_delivered_in_watt,
        )
        stsv.set_output_value(self.heating_channel, self.state.heating_power_in_watt)
        stsv.set_output_value(self.cooling_channel, self.state.cooling_power_in_watt)
        stsv.set_output_value(self.electricity_output_channel, self.state.electricity_input_in_watt)
        stsv.set_output_value(self.number_of_cycles_channel, self.number_of_cycles)

    def process_thermal(self, ws_in: float) -> None:
        """Process thermal."""
        pass
        # temperature_max = 55
        # heat_capacity = PhysicsConfig.water_specific_heat_capacity
        # thermal_energy_to_add = self.max_heating_power
        # ws_out_mass = ws_in.mass
        # try:
        # ws_out_temperature = ws_in.temperature + thermal_energy_to_add / (heat_capacity * ws_out_mass)
        # except ZeroDivisionError:
        # log.information(heat_capacity)
        # log.information(ws_out_mass)
        # log.information(ws_in.mass)
        # raise ValueError
        # wasted_energy = 0
        # if ws_out_temperature > temperature_max:
        # delta_T = ws_out_temperature - temperature_max
        # wasted_energy = (delta_T * ws_out_mass * PhysicsConfig.water_specific_heat_capacity)
        #  ws_out_temperature = temperature_max
        # ws_out_enthalpy = ws_in.enthalpy + thermal_energy_to_add
        # ws_in.change_slice_parameters(new_temperature=ws_out_temperature, new_enthalpy=ws_out_enthalpy, new_mass=ws_out_mass)
        # return ws_in, wasted_energy, thermal_energy_to_add

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of electricity costs and revenues."""
        opex_cost_data_class = OpexCostDataClass.get_default_opex_cost_data_class()
        return opex_cost_data_class

    @staticmethod
    def get_cost_capex(config: GenericHeatPumpConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:  # pylint: disable=unused-argument
        """Returns investment cost, CO2 emissions and lifetime."""
        capex_cost_data_class = CapexCostDataClass.get_default_capex_cost_data_class()
        return capex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        return []


class GenericHeatPumpController(cp.Component):
    """Heat Pump Controller.

    It takes data from other
    components and sends signal to the heat pump for
    activation or deactivation.

    Parameters
    ----------
    t_air_heating: float
        Minimum comfortable temperature for residents
    t_air_cooling: float
        Maximum comfortable temperature for residents
    offset: float
        Temperature offset to compensate the hysteresis
        correction for the building temperature change
    mode : int
        Mode index for operation type for this heat pump

    """

    # Inputs
    TemperatureMean = "Residence Temperature"
    ElectricityInput = "ElectricityInput"

    # Outputs
    State = "State"

    # Similar components to connect to:
    # 1. Building
    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: GenericHeatPumpControllerConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Construct all the neccessary attributes."""
        self.heatpump_controller_config = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.build(
            temperature_air_cooling=self.heatpump_controller_config.temperature_air_cooling_in_celsius,
            temperature_air_heating=self.heatpump_controller_config.temperature_air_heating_in_celsius,
            offset=self.heatpump_controller_config.offset,
            mode=self.heatpump_controller_config.mode,
        )

        self.temperature_mean_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureMean,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            True,
        )
        self.electricity_input_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ElectricityInput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            False,
        )
        self.state_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.State,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description=f"here a description for {self.State} will follow.",
        )

        self.add_default_connections(self.get_default_connections_from_building())
        self.add_default_connections(self.get_default_connections_from_electricity_meter())
        self.controller_heatpumpmode: Any
        self.previous_heatpump_mode: Any

    def get_default_connections_from_building(self) -> List[cp.ComponentConnection]:
        """Get building default connections."""

        connections = []
        building_classname = Building.get_classname()
        connections.append(
            cp.ComponentConnection(
                GenericHeatPumpController.TemperatureMean,
                building_classname,
                Building.TemperatureMeanThermalMass,
            )
        )
        return connections

    def get_default_connections_from_electricity_meter(self) -> List[cp.ComponentConnection]:
        """Get electricity meter default connections."""

        from hisim.components.electricity_meter import ElectricityMeter  # pylint: disable=import-outside-toplevel

        connections = []
        em_classname = ElectricityMeter.get_classname()
        connections.append(
            cp.ComponentConnection(
                GenericHeatPumpController.ElectricityInput,
                em_classname,
                ElectricityMeter.ElectricityAvailable,
            )
        )
        return connections

    def build(
        self,
        temperature_air_heating: float,
        temperature_air_cooling: float,
        offset: float,
        mode: float,
    ) -> None:
        """Build function.

        The function sets important constants and parameters for the calculations.
        """
        # Sth
        self.controller_heatpumpmode = "off"
        self.previous_heatpump_mode = self.controller_heatpumpmode

        # Configuration
        self.temperature_set_heating = temperature_air_heating
        self.temperature_set_cooling = temperature_air_cooling
        self.offset = offset

        self.mode = mode

    def i_prepare_simulation(self) -> None:
        """Prepare the simulation."""
        pass

    def i_save_state(self) -> None:
        """Save the current state."""
        self.previous_heatpump_mode = self.controller_heatpumpmode

    def i_restore_state(self) -> None:
        """Restore the previous state."""
        self.controller_heatpumpmode = self.previous_heatpump_mode

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublecheck."""
        pass

    def write_to_report(self) -> List[str]:
        """Write important variables to report."""
        return self.heatpump_controller_config.get_string_dict()

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the heat pump comtroller."""
        # check demand, and change state of self.has_heating_demand, and self._has_cooling_demand
        if force_convergence:
            pass
        else:
            # Retrieves inputs
            temperature_mean_old = stsv.get_input_value(self.temperature_mean_channel)
            electricity_input = stsv.get_input_value(self.electricity_input_channel)

            if self.mode == 1:
                self.conditions(temperature_mean_old)
            elif self.mode == 2:
                self.smart_conditions(temperature_mean_old, electricity_input)

        if self.controller_heatpumpmode == "heating":
            state = 1
        if self.controller_heatpumpmode == "cooling":
            state = -1
        if self.controller_heatpumpmode == "off":
            state = 0
        stsv.set_output_value(self.state_channel, state)

    def conditions(self, set_temperature: float) -> None:
        """Set conditions for the heat pump controller mode."""
        maximum_heating_set_temperature = self.temperature_set_heating + self.offset
        minimum_heating_set_temperature = self.temperature_set_heating
        minimum_cooling_set_temperature = self.temperature_set_cooling - self.offset
        maximum_cooling_set_temperature = self.temperature_set_cooling

        if self.controller_heatpumpmode == "heating":  # and daily_avg_temp < 15:
            if set_temperature > maximum_heating_set_temperature:  # 23
                self.controller_heatpumpmode = "off"
                return
        if self.controller_heatpumpmode == "cooling":
            if set_temperature < minimum_cooling_set_temperature:  # 24
                self.controller_heatpumpmode = "off"
                return
        if self.controller_heatpumpmode == "off":
            # if pvs_surplus > ? and air_temp < minimum_heating_air + 2:
            if set_temperature < minimum_heating_set_temperature:  # 21
                self.controller_heatpumpmode = "heating"
                return
            if set_temperature > maximum_cooling_set_temperature:  # 26
                self.controller_heatpumpmode = "cooling"
                return

    def smart_conditions(self, set_temperature: float, electricity_input: float) -> None:
        """Set smart conditions for the heat pump controller mode."""
        smart_offset_upper = 3
        smart_offset_lower = 0.5
        maximum_heating_set_temperature = self.temperature_set_heating + self.offset
        if electricity_input < 0:
            maximum_heating_set_temperature += smart_offset_upper
        # maximum_heating_set_temp = self.t_set_heating
        minimum_heating_set_temperature = self.temperature_set_heating
        if electricity_input < 0:
            minimum_heating_set_temperature += smart_offset_lower
        minimum_cooling_set_temperature = self.temperature_set_cooling - self.offset
        # minimum_cooling_set_temp = self.t_set_cooling
        maximum_cooling_set_temperature = self.temperature_set_cooling

        if self.controller_heatpumpmode == "heating":  # and daily_avg_temp < 15:
            if set_temperature > maximum_heating_set_temperature:  # 23
                self.controller_heatpumpmode = "off"
                return
        if self.controller_heatpumpmode == "cooling":
            if set_temperature < minimum_cooling_set_temperature:  # 24
                self.controller_heatpumpmode = "off"
                return
        if self.controller_heatpumpmode == "off":
            # if pvs_surplus > ? and air_temp < minimum_heating_air + 2:
            if set_temperature < minimum_heating_set_temperature:  # 21
                self.controller_heatpumpmode = "heating"
                return
            if set_temperature > maximum_cooling_set_temperature:  # 26
                self.controller_heatpumpmode = "cooling"
                return

        # if timestep >= 60*24*30*3 and timestep <= 60*24*30*9:  #    state = 0

        # log.information("Final state: {}\n".format(state))

    def prin1t_outpu1t(self, t_m: float, state: Any) -> None:
        """Print output of heat pump controller."""
        log.information("==========================================")
        log.information(f"T m: {t_m}")
        log.information(f"State: {state}")

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of electricity costs and revenues."""
        opex_cost_data_class = OpexCostDataClass.get_default_opex_cost_data_class()
        return opex_cost_data_class

    @staticmethod
    def get_cost_capex(config: GenericHeatPumpControllerConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:  # pylint: disable=unused-argument
        """Returns investment cost, CO2 emissions and lifetime."""
        capex_cost_data_class = CapexCostDataClass.get_default_capex_cost_data_class()
        return capex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        return []



""" L2 Controller for PtX Buffer Battery operation. """

# clean
import os
from typing import List, Any
import json
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from hisim.component import ConfigBase, Component, ComponentInput, ComponentOutput, SingleTimeStepValues, DisplayConfig

from hisim import loadtypes as lt
from hisim import utils
from hisim.simulationparameters import SimulationParameters

__authors__ = "Franz Oldopp"
__copyright__ = "Copyright 2023, IEK-3"
__credits__ = ["Franz Oldopp"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Franz Oldopp"
__email__ = "f.oldopp@fz-juelich.de"
__status__ = "development"


@dataclass_json
@dataclass
class XTPControllerConfig(ConfigBase):
    """Configutation of the PtX  Controller."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return XTPController.get_full_classname()

    building_name: str
    name: str
    nom_output: float
    min_output: float
    max_output: float
    standby_load: float
    operation_mode: str

    @staticmethod
    def read_config(fuel_cell_name):
        """Read config."""
        config_file = os.path.join(utils.HISIMPATH["inputs"], "fuel_cell_manufacturer_config.json")
        with open(config_file, "r", encoding="utf-8") as json_file:
            data = json.load(json_file)
            return data.get("Fuel Cell variants", {}).get(fuel_cell_name, {})

    @classmethod
    def control_fuel_cell(
        cls,
        fuel_cell_name: str,
        operation_mode: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Sets the according parameters for the chosen fuel cell."""
        config_json = cls.read_config(fuel_cell_name)

        config = XTPControllerConfig(
            building_name=building_name,
            name="L2XTPController",  # config_json.get("name", "")
            nom_output=config_json.get("nom_output", 0.0),
            min_output=config_json.get("min_output", 0.0),
            max_output=config_json.get("max_output", 0.0),
            standby_load=config_json.get("standby_load", 0.0),
            operation_mode=operation_mode,
        )
        return config


class XTPController(Component):
    """XtP  Controller."""

    # Inputs
    DemandLoad = "DemandLoad"
    StateOfCharge = "StateOfCharge"

    # Outputs
    PowerFromThird = "PowerFromThird"
    DemandToSystem = "DemandToSystem"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: XTPControllerConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.xtpcontrollerconfig = config

        self.nom_output = config.nom_output
        self.min_output = config.min_output
        self.max_output = config.max_output
        self.standby_load = config.standby_load
        self.operation_mode = config.operation_mode

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        # =================================================================================================================================
        # Input channels

        self.demand_input: ComponentInput = self.add_input(
            self.component_name,
            XTPController.DemandLoad,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            True,
        )

        self.soc: ComponentInput = self.add_input(
            self.component_name,
            XTPController.StateOfCharge,
            lt.LoadTypes.DISTRICTHEATING,
            lt.Units.PERCENT,
            False,
        )

        # =================================================================================================================================
        # Output channels

        self.load_from_battery: ComponentOutput = self.add_output(
            self.component_name,
            XTPController.PowerFromThird,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            output_description="Discharges the battery in case of no power production",
        )

        self.demand_to_system: ComponentOutput = self.add_output(
            self.component_name,
            XTPController.DemandToSystem,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            output_description="distributes demand to the system",
        )

        # =================================================================================================================================
        # Initialize variables
        self.system_state = "OFF"
        self.threshold_exceeded = False
        self.standby_time_count = 0.0

        self.system_state_previous = self.system_state
        self.threshold_exceeded_previous = self.threshold_exceeded
        self.standby_time_count_previous = self.standby_time_count

    def system_operation(self, operation_mode, demand_load):
        """System operation."""
        if operation_mode == "StandbyLoad":
            if self.min_output <= demand_load <= self.max_output:
                demand_to_system = demand_load
                power_from_battery = 0.0
            elif self.max_output < demand_load:
                demand_to_system = demand_load
                power_from_battery = 0.0
            elif self.standby_load <= demand_load < self.min_output:
                # standby_load <= demand_load < min_output and demand_load < standby_load:
                demand_to_system = demand_load
                power_from_battery = 0.0
            else:
                demand_to_system = self.standby_load
                power_from_battery = -self.standby_load

        elif operation_mode == "StandbyandOffLoad":
            if self.min_output <= demand_load <= self.max_output:
                self.standby_time_count = 0.0
                demand_to_system = demand_load
                power_from_battery = 0.0
            elif self.max_output < demand_load:
                self.standby_time_count = 0.0
                demand_to_system = demand_load
                power_from_battery = 0.0
            elif self.standby_load <= demand_load < self.min_output:
                self.standby_time_count = 0.0
                demand_to_system = demand_load
                power_from_battery = 0.0
            else:  # demand_load <= self.standby_load
                standby_operation_time = 7200.0  # 7200.0
                if self.standby_time_count >= standby_operation_time:
                    demand_to_system = 0.0
                    power_from_battery = 0.0
                else:
                    demand_to_system = self.standby_load
                    power_from_battery = -self.standby_load
                    self.standby_time_count += self.my_simulation_parameters.seconds_per_timestep

        else:
            demand_to_system = demand_load
            power_from_battery = 0.0

        return demand_to_system, power_from_battery

    def i_prepare_simulation(self) -> None:
        """Prepare the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        self.system_state_previous = self.system_state
        self.threshold_exceeded_previous = self.threshold_exceeded
        self.standby_time_count_previous = self.standby_time_count

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.system_state = self.system_state_previous
        self.threshold_exceeded = self.threshold_exceeded_previous
        self.standby_time_count = self.standby_time_count_previous

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the component."""
        if force_convergence:
            return

        demand_load = abs(stsv.get_input_value(self.demand_input) / 1000)  # WATT input to KILOWATT

        """ Only for household testing
        if self.system_state == "OFF" and soc < 0.2:
            power_from_battery = demand_load
            demand_to_system = 0.0
        elif self.system_state == "OFF" and soc >= 0.2:
            (demand_to_system, power_from_battery) = self.system_operation(
                self.operation_mode, demand_load
            )
            self.system_state = "ON"
        else:
            (demand_to_system, power_from_battery) = self.system_operation(
                self.operation_mode, demand_load
            )
            self.system_state = "ON"
        """
        (demand_to_system, power_from_battery) = self.system_operation(self.operation_mode, demand_load)

        """
        if self.system_state == "OFF":
            if 0.30 < stsv.get_input_value(self.soc):
                print(stsv.get_input_value(self.soc))
                self.system_state = "ON"
                print(self.system_state)

            power_from_battery = demand_load
            demand_to_system = 0.0
        """

        stsv.set_output_value(self.load_from_battery, power_from_battery * 1000)  # Battery Output in WATT
        stsv.set_output_value(self.demand_to_system, demand_to_system)

    def write_to_report(self) -> List[str]:
        """Writes a report."""
        return self.xtpcontrollerconfig.get_string_dict()



"""Import weather data from dwd and era5."""

import os
import urllib.request as url
import datetime
import pandas as pd
from wetterdienst import Settings
from wetterdienst.provider.dwd.observation import (
    DwdObservationParameter,
    DwdObservationRequest,
    DwdObservationResolution,
)
import numpy as np
import cdsapi
import xarray as xr
from hisim.components.weather import WeatherDataSourceEnum


__authors__ = "Jonas Hoppe"
__copyright__ = ""
__credits__ = [" Jonas Hoppe "]
__license__ = ""
__version__ = ""
__maintainer__ = "  "
__email__ = ""
__status__ = ""


class WeatherDataImport:
    """Import class for weather data from dwd and era5."""

    def __init__(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        location: str,
        latitude: float,
        longitude: float,
        distance_weather_stations: float,
        path_input_folder: str,
        weather_data_source: WeatherDataSourceEnum,
    ):
        """Initializes Data Request."""
        self.location = location
        self.latitude = latitude
        self.longitude = longitude
        self.distance_weather_stations = distance_weather_stations
        self.path_input_folder = path_input_folder
        self.weather_data_source = weather_data_source

        self.long_round = round(self.longitude, 2)
        self.lat_round = round(self.latitude, 2)

        location_data = [
            {
                "location": self.location,
                "longitude": self.longitude,
                "latitude": self.latitude,
            }
        ]
        self.location_df = pd.DataFrame(location_data)

        start_year = start_date.year

        if end_date.year == start_date.year:
            end_year = end_date.year + 1
        elif end_date.year > start_date.year:
            end_year = end_date.year
        else:
            raise KeyError("Check start and end date of weather data request!")

        self.start_date_for_weather_data = datetime.datetime(
            year=start_year, month=1, day=1, hour=0, minute=0, second=0
        )
        self.end_date_for_weather_data = datetime.datetime(year=end_year, month=1, day=1, hour=0, minute=0, second=0)

        print(
            f"Weatherdata request from {self.start_date_for_weather_data} to {self.end_date_for_weather_data} in {weather_data_source} database."
        )

        self.csv_path: str

        if self.weather_data_source == WeatherDataSourceEnum.DWD_10MIN:
            self.dwd_10min_request()
        elif self.weather_data_source == WeatherDataSourceEnum.ERA5:
            self.era5_request()
        else:
            raise KeyError("Unknown data source. Only DWD_10MIN and ERA5.")

    def safe_as_csv(self, weather_df, csv_path):
        """Safe weather data as .csv in given path."""
        self.location_df.to_csv(csv_path, index=False)

        weather_df.to_csv(csv_path, mode="a", index=False)

        print(f"Weather data saved in:  {csv_path}")

    def dwd_10min_request(
        self,
    ):
        """Data request from "Deutschen Wetterdienst".

        Based on:
        https://wetterdienst.readthedocs.io/en/latest/data/coverage/dwd/observation/minute_10.html

        Weighted average based on distance --> http://www.hydrology.uni-freiburg.de/abschluss/Niederberger_J_2000_DA.pdf .
        """

        csv_name = f"dwd_10min_{self.start_date_for_weather_data.year}-{self.end_date_for_weather_data.year}_{self.location}_{self.long_round}_{self.lat_round}.csv"
        speicherpfad = os.path.join(self.path_input_folder, "weather", "dwd_10min_data")

        # Abfrage, ob csv schon vorhanden

        if not os.path.exists(speicherpfad):
            os.makedirs(speicherpfad)
            print(f"Directory {speicherpfad} created.")

        csv_path = os.path.join(speicherpfad, csv_name)

        if os.path.isfile(csv_path):
            print(f"CSV file  {csv_name} under  {speicherpfad} already exists.")
        else:
            print(f"Weather data is loaded from DWD and saved under {csv_path}.")

            settings = Settings(  # default
                ts_shape="long",  # tidy data
                ts_humanize=True,  # humanized parameters
                ts_si_units=False,  # convert values to SI units
                ts_interpolation_use_nearby_station_distance=40,
            )

            parameters = [DwdObservationParameter.MINUTE_10.TEMPERATURE_AIR_MEAN_2M,
                          DwdObservationParameter.MINUTE_10.PRESSURE_AIR_SITE,
                          DwdObservationParameter.MINUTE_10.WIND_DIRECTION,
                          DwdObservationParameter.MINUTE_10.WIND_SPEED,
                          DwdObservationParameter.MINUTE_10.RADIATION_SKY_SHORT_WAVE_DIFFUSE,
                          DwdObservationParameter.MINUTE_10.RADIATION_GLOBAL,
                          ]

            request = DwdObservationRequest(
                parameter=parameters,  # ["tt_10", "pp_10", "dd_10", "ff_10", "ds_10", "gs_10"],
                resolution=DwdObservationResolution.MINUTE_10,  # "10_minutes",
                start_date=str(self.start_date_for_weather_data),
                end_date=str(self.end_date_for_weather_data),
                settings=settings,
            ).filter_by_distance(
                latlon=(self.latitude, self.longitude),
                distance=self.distance_weather_stations,
                unit="km",
            )

            stations = request.df

            if len(stations) == 0:
                raise KeyError("No weather station found in the area. Increase radius for station search.")
            if len(stations) < 3:
                raise KeyError(
                    "Only 2 weather stations found in the area. Minimum number 3! Increase radius for station search."
                )

            stations = stations.to_pandas()
            distance_location_to_stations_df = pd.DataFrame(
                {
                    "longitude": stations["longitude"],
                    "latitude": stations["latitude"],
                    "distance": stations["distance"],
                }
            )

            # Calculating weight based on inverse distance
            radial_distance = np.sqrt(
                np.square(self.longitude - distance_location_to_stations_df["longitude"])
                + np.square(self.latitude - distance_location_to_stations_df["latitude"])
            )
            weights = 1 / (np.sqrt(radial_distance))
            weights = pd.DataFrame({"weights": weights})

            values = request.values.all().df
            values = values.to_pandas()

            parameterlist = set()
            for element in values["parameter"]:
                parameterlist.add(element)
            print(f"DWD Parameter names: {parameterlist}")

            date_df = pd.DataFrame(columns=["date"])
            date_df = pd.to_datetime(values["date"], utc=True)
            date_df = pd.DataFrame({"date": date_df})
            date_df = date_df["date"].drop_duplicates()

            time_df = pd.DataFrame(columns=["year", "month", "day", "hour", "minute"])

            time_df["year"] = date_df.dt.year
            time_df["month"] = date_df.dt.month
            time_df["day"] = date_df.dt.day
            time_df["hour"] = date_df.dt.hour
            time_df["minute"] = date_df.dt.minute

            missing_time_df = pd.DataFrame()
            if len(time_df) == 60 * 24 * 365 / 10 + 1:
                print("All weather data for a year is available.")
            else:
                if 60 * 24 * 365 / 10 - len(time_df) + 1 < 50:
                    print(f"Note: {60 * 24 * 365 /10 - len(time_df)+1} entries for an entire year are missing.")

                    time_index_full = pd.to_datetime(time_df[["year", "month", "day", "hour", "minute"]])

                    missing_times = pd.date_range(
                        start=time_index_full.min(),
                        end=time_index_full.max(),
                        freq="1H",
                    ).difference(time_index_full)

                    missing_times = pd.to_datetime(missing_times)

                    missing_times_df_year = pd.DataFrame({"year": missing_times.year})
                    missing_times_df_month = pd.DataFrame({"month": missing_times.month})
                    missing_times_df_day = pd.DataFrame({"day": missing_times.day})
                    missing_times_df_hour = pd.DataFrame({"hour": missing_times.hour})
                    missing_times_df_minute = pd.DataFrame({"minute": missing_times.minute})

                    missing_time_df = pd.concat(
                        [
                            missing_times_df_year["year"],
                            missing_times_df_month["month"],
                            missing_times_df_day["day"],
                            missing_times_df_hour["hour"],
                            missing_times_df_minute["minute"],
                        ],
                        axis=1,
                    )

                    print("Missing timeseries:")
                    print(missing_times)

                else:
                    raise KeyError(
                        f"Note: {60 * 24 * 365 /10 - len(time_df)+1} entries for an entire year are missing. Too much!"
                    )
            print("Write Weather Data into Dataframe.")
            temperature_dwd_df = (
                values[values["parameter"] == "temperature_air_mean_2m"]
                .groupby("date")["value"]
                .apply(lambda x: ", ".join(map(str, x)))
                .reset_index()
            )
            temperature_dwd_df = temperature_dwd_df.rename(columns={"value": "temperature"})
            temperature_dwd_df["temperature"] = temperature_dwd_df["temperature"].apply(
                lambda x: [float(val) if val != "nan" else np.nan for val in x.split(", ")]
            )

            temperatur_air_list = []
            temperature_air_valid_data = pd.DataFrame()
            for i in range(len(temperature_dwd_df)):
                temperature_air_valid_data = pd.DataFrame(
                    {"temperature": temperature_dwd_df["temperature"][i]}
                ).dropna()

                if temperature_air_valid_data.empty:
                    temperature_location = np.nan
                else:
                    temperature_location = sum(
                        temperature_air_valid_data["temperature"] * weights["weights"][temperature_air_valid_data.index]
                    ) / sum(weights["weights"][temperature_air_valid_data.index])

                temperatur_air_list.append(temperature_location)

            temperature_air_df = pd.DataFrame(temperatur_air_list, columns=["temperature"])
            temperature_air_df = temperature_air_df.interpolate(method="linear", limit_direction="backward")

            if temperature_air_df.empty:
                raise KeyError("temperature_air_df is empty --> error in weather data request. Change location")

            pressure_dwd_df = (
                values[values["parameter"] == "pressure_air_site"]
                .groupby("date")["value"]
                .apply(lambda x: ", ".join(map(str, x)))
                .reset_index()
            )
            pressure_dwd_df = pressure_dwd_df.rename(columns={"value": "pressure"})
            pressure_dwd_df["pressure"] = pressure_dwd_df["pressure"].apply(
                lambda x: [float(val) if val != "nan" else np.nan for val in x.split(", ")]
            )

            pressure_list = []
            pressure_valid_data = pd.DataFrame()
            for i in range(len(temperature_dwd_df)):
                pressure_valid_data = pd.DataFrame({"pressure": pressure_dwd_df["pressure"][i]}).dropna()

                if pressure_valid_data.empty:
                    pressure_location = np.nan
                else:
                    pressure_location = sum(
                        pressure_valid_data["pressure"] * weights["weights"][pressure_valid_data.index]
                    ) / sum(weights["weights"][pressure_valid_data.index])

                pressure_list.append(pressure_location)

            pressure_df = pd.DataFrame(pressure_list, columns=["pressure"])
            pressure_df = pressure_df.interpolate(method="linear", limit_direction="backward")

            if pressure_df.empty:
                raise KeyError("pressure_df is empty --> error in weather data request. Change location")

            wind_direction_dwd_df = (
                values[values["parameter"] == "wind_direction"]
                .groupby("date")["value"]
                .apply(lambda x: ", ".join(map(str, x)))
                .reset_index()
            )
            wind_direction_dwd_df = wind_direction_dwd_df.rename(columns={"value": "wind_direction"})
            wind_direction_dwd_df["wind_direction"] = wind_direction_dwd_df["wind_direction"].apply(
                lambda x: [float(val) if val != "nan" else np.nan for val in x.split(", ")]
            )

            wind_direction_list = []
            wind_direction_valid_data = pd.DataFrame()
            for i in range(len(wind_direction_dwd_df)):
                wind_direction_valid_data = pd.DataFrame(
                    {"wind_direction": wind_direction_dwd_df["wind_direction"][i]}
                ).dropna()

                if wind_direction_valid_data.empty:
                    wind_direction_location = np.nan
                else:
                    wind_direction_location = sum(
                        wind_direction_valid_data["wind_direction"]
                        * weights["weights"][wind_direction_valid_data.index]
                    ) / sum(weights["weights"][wind_direction_valid_data.index])

                wind_direction_list.append(wind_direction_location)

            wind_direction_df = pd.DataFrame(wind_direction_list, columns=["wind_direction"])
            wind_direction_df = wind_direction_df.interpolate(method="linear", limit_direction="backward")

            if wind_direction_df.empty:
                raise KeyError("wind_direction_df is empty --> error in weather data request. Change location")

            wind_speed_dwd_df = (
                values[values["parameter"] == "wind_speed"]
                .groupby("date")["value"]
                .apply(lambda x: ", ".join(map(str, x)))
                .reset_index()
            )
            wind_speed_dwd_df = wind_speed_dwd_df.rename(columns={"value": "wind_speed"})
            wind_speed_dwd_df["wind_speed"] = wind_speed_dwd_df["wind_speed"].apply(
                lambda x: [float(val) if val != "nan" else np.nan for val in x.split(", ")]
            )

            wind_speed_list = []
            wind_speed_valid_data = pd.DataFrame()
            for i in range(len(wind_speed_dwd_df)):
                wind_speed_valid_data = pd.DataFrame({"wind_speed": wind_speed_dwd_df["wind_speed"][i]}).dropna()

                if wind_speed_valid_data.empty:
                    wind_speed_location = np.nan
                else:
                    wind_speed_location = sum(
                        wind_speed_valid_data["wind_speed"] * weights["weights"][wind_speed_valid_data.index]
                    ) / sum(weights["weights"][wind_speed_valid_data.index])

                wind_speed_list.append(wind_speed_location)

            wind_speed_df = pd.DataFrame(wind_speed_list, columns=["wind_speed"])
            wind_speed_df = wind_speed_df.interpolate(method="linear", limit_direction="backward")

            if wind_speed_df.empty:
                raise KeyError("wind_speed_df is empty --> error in weather data request. Change location")

            diffuse_irradiance_dwd_df = (
                values[values["parameter"] == "radiation_sky_short_wave_diffuse"]
                .groupby("date")["value"]
                .apply(lambda x: ", ".join(map(str, x)))
                .reset_index()
            )
            diffuse_irradiance_dwd_df = diffuse_irradiance_dwd_df.rename(columns={"value": "diffuse_irradiance"})
            diffuse_irradiance_dwd_df["diffuse_irradiance"] = diffuse_irradiance_dwd_df["diffuse_irradiance"].apply(
                lambda x: [float(val) if val != "nan" else np.nan for val in x.split(", ")]
            )

            diffuse_irradiance_list = []
            diffuse_irradiance_valid_data = pd.DataFrame()
            for i in range(len(diffuse_irradiance_dwd_df)):
                diffuse_irradiance_valid_data = pd.DataFrame(
                    {"diffuse_irradiance": diffuse_irradiance_dwd_df["diffuse_irradiance"][i]}
                ).dropna()

                if diffuse_irradiance_valid_data.empty:
                    diffuse_irradiance_location_watt_per_m2 = np.nan
                else:
                    diffuse_irradiance_location_j_per_cm2 = sum(
                        diffuse_irradiance_valid_data["diffuse_irradiance"]
                        * weights["weights"][diffuse_irradiance_valid_data.index]
                    ) / sum(weights["weights"][diffuse_irradiance_valid_data.index])

                    diffuse_irradiance_location_watt_per_m2 = diffuse_irradiance_location_j_per_cm2 * 10000 / (10 * 60)

                diffuse_irradiance_list.append(diffuse_irradiance_location_watt_per_m2)

            diffuse_irradiance_df = pd.DataFrame(diffuse_irradiance_list, columns=["diffuse_irradiance"])
            diffuse_irradiance_df = diffuse_irradiance_df.interpolate(method="linear", limit_direction="backward")

            if diffuse_irradiance_df.empty:
                raise KeyError("diffuse_irradiance_df is empty --> error in weather data request. Change location")

            global_irradiance_dwd_df = (
                values[values["parameter"] == "radiation_global"]
                .groupby("date")["value"]
                .apply(lambda x: ", ".join(map(str, x)))
                .reset_index()
            )
            global_irradiance_dwd_df = global_irradiance_dwd_df.rename(columns={"value": "global_irradiance"})
            global_irradiance_dwd_df["global_irradiance"] = global_irradiance_dwd_df["global_irradiance"].apply(
                lambda x: [float(val) if val != "nan" else np.nan for val in x.split(", ")]
            )

            global_irradiance_list = []
            global_irradiance_valid_data = pd.DataFrame()
            for i in range(len(global_irradiance_dwd_df)):
                global_irradiance_valid_data = pd.DataFrame(
                    {"global_irradiance": global_irradiance_dwd_df["global_irradiance"][i]}
                ).dropna()

                if global_irradiance_valid_data.empty:
                    global_irradiance_location_watt_per_m2 = np.nan
                else:
                    global_irradiance_location_j_per_cm2 = sum(
                        global_irradiance_valid_data["global_irradiance"]
                        * weights["weights"][global_irradiance_valid_data.index]
                    ) / sum(weights["weights"][global_irradiance_valid_data.index])

                    global_irradiance_location_watt_per_m2 = global_irradiance_location_j_per_cm2 * 10000 / (10 * 60)

                global_irradiance_list.append(global_irradiance_location_watt_per_m2)

            global_irradiance_df = pd.DataFrame(global_irradiance_list, columns=["global_irradiance"])
            global_irradiance_df = global_irradiance_df.interpolate(method="linear", limit_direction="backward")

            if global_irradiance_df.empty:
                raise KeyError("global_irradiance_df is empty --> error in weather data request. Change location")

            weather_df = pd.DataFrame()
            weather_df = pd.concat(
                [
                    time_df,
                    temperature_air_df,  # ["temperature"],
                    pressure_df,  # ["pressure"],
                    wind_direction_df,  # ["wind_direction"],
                    wind_speed_df,  # ["wind_speed"],
                    diffuse_irradiance_df,  # ["diffuse_irradiance"],
                    global_irradiance_df,  # ["global_irradiance"],
                ],
                axis=1,
                join="outer",
            )

            if len(missing_time_df) != 0:
                # Setze den Zeitindex für df2
                weather_df_time_index = pd.to_datetime(weather_df[["year", "month", "day", "hour", "minute"]])
                time_index_full = pd.to_datetime(missing_time_df[["year", "month", "day", "hour", "minute"]])

                merged_weather_df = pd.concat(
                    [
                        missing_time_df.set_index(time_index_full),
                        weather_df.set_index(weather_df_time_index),
                    ]
                ).sort_index()

                interpolated_weather_df = merged_weather_df.interpolate()

                weather_df = interpolated_weather_df
                print("Missing times successfully added and weather data interpolated!")

            weather_df = weather_df.reset_index(drop=True)
            weather_df = weather_df.drop(weather_df.index[-1])

            # if weather_df.isna().any().any():
            #     raise KeyError(
            #         "The DataFrame weather_df contains NaN values. Check data or increase radius for station search."
            #     )
            #
            # if weather_df.eq("nan").any().any():
            #     raise KeyError(
            #         "The DataFrame weather_df contains NaN values. Check data or increase radius for station search."
            #     )

            self.safe_as_csv(weather_df, csv_path)
        print("Safe Weatherdata successfully in .csv.")
        self.csv_path = csv_path

    def era5_request(
        self,
    ):
        """Data request from era5.

        Based on:
        https://confluence.ecmwf.int/display/CKB/ERA5%3A+data+documentation
        https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels?tab=overview
        https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels?tab=form

        warning: Necessary to set up an account and match URL and user key to storage for retrieval.
        """

        # Check if .cdsapirc key for data request is available
        username = os.getlogin()
        user_path = os.path.join("C:\\Users", username)

        file_name = ".cdsapirc"
        file_path = os.path.join(user_path, file_name)

        if not os.path.isfile(file_path):
            raise KeyError(
                f"The config-file {file_name} for era5 data request is missing in {user_path}. "
                f"Please register: https://cds.climate.copernicus.eu/user/login "
                f"and follow the instructions for using era5."
            )

        csv_name = f"era5_1h_{self.start_date_for_weather_data.year}-{self.end_date_for_weather_data.year}_{self.location}_{self.long_round}_{self.lat_round}.csv"
        speicherpfad = os.path.join(self.path_input_folder, "weather", "era5_data")

        if not os.path.exists(speicherpfad):
            os.makedirs(speicherpfad)
            print(f"Directory {speicherpfad} created.")

        csv_path = os.path.join(speicherpfad, csv_name)

        if os.path.isfile(csv_path):
            print(f"CSV file  {csv_name} under  {speicherpfad} already exists.")
        else:
            print(f"Weather data is loaded from DWD and saved under {csv_path}.")

            cds = cdsapi.Client()

            data = cds.retrieve(
                "reanalysis-era5-single-levels",
                {
                    "product_type": "reanalysis",
                    "variable": [
                        "2m_temperature",
                        "10m_u_component_of_wind",
                        "10m_v_component_of_wind",
                        "surface_pressure",
                        "surface_solar_radiation_downwards",
                        "total_sky_direct_solar_radiation_at_surface",
                    ],
                    "year": [
                        str(self.start_date_for_weather_data.year),
                    ],
                    "month": [
                        "01",
                        "02",
                        "03",
                        "04",
                        "05",
                        "06",
                        "07",
                        "08",
                        "09",
                        "10",
                        "11",
                        "12",
                    ],
                    "day": [
                        "01",
                        "02",
                        "03",
                        "04",
                        "05",
                        "06",
                        "07",
                        "08",
                        "09",
                        "10",
                        "11",
                        "12",
                        "13",
                        "14",
                        "15",
                        "16",
                        "17",
                        "18",
                        "19",
                        "20",
                        "21",
                        "22",
                        "23",
                        "24",
                        "25",
                        "26",
                        "27",
                        "28",
                        "29",
                        "30",
                        "31",
                    ],
                    "time": [
                        "00:00",
                        "01:00",
                        "02:00",
                        "03:00",
                        "04:00",
                        "05:00",
                        "06:00",
                        "07:00",
                        "08:00",
                        "09:00",
                        "10:00",
                        "11:00",
                        "12:00",
                        "13:00",
                        "14:00",
                        "15:00",
                        "16:00",
                        "17:00",
                        "18:00",
                        "19:00",
                        "20:00",
                        "21:00",
                        "22:00",
                        "23:00",
                    ],
                    "area": [
                        self.latitude,
                        self.longitude,
                        self.latitude,
                        self.longitude,
                    ],
                    "format": "netcdf",
                },
                # 'download.nc'
            )

            with url.urlopen(data.location) as data_set:
                weather_data_set = xr.open_dataset(data_set.read())
            values = weather_data_set.to_dataframe()

            values = values.dropna()

            time_value = values.index.get_level_values("time")
            df_time = pd.DataFrame(index=time_value, columns=["data_column"])
            df_year = pd.DataFrame({"year": df_time.index.year})
            df_month = pd.DataFrame({"month": df_time.index.month})
            df_day = pd.DataFrame({"day": df_time.index.day})
            df_hour = pd.DataFrame({"hour": df_time.index.hour})
            df_minute = pd.DataFrame({"minute": df_time.index.minute})

            time_df = pd.concat(
                [
                    df_year["year"],
                    df_month["month"],
                    df_day["day"],
                    df_hour["hour"],
                    df_minute["minute"],
                ],
                axis=1,
            )

            missing_time_df = pd.DataFrame()

            if len(time_df) == 24 * 365:
                print("All weather data for a year is available.")
            else:
                if 24 * 365 - len(time_df) < 50:
                    print(f"Note: {24 * 365 - len(time_df)} entries for an entire year are missing.")

                    time_index_full = pd.to_datetime(time_df[["year", "month", "day", "hour", "minute"]])

                    missing_times = pd.date_range(
                        start=time_index_full.min(),
                        end=time_index_full.max(),
                        freq="1H",
                    ).difference(time_index_full)

                    missing_times = pd.to_datetime(missing_times)

                    missing_times_df_year = pd.DataFrame({"year": missing_times.year})
                    missing_times_df_month = pd.DataFrame({"month": missing_times.month})
                    missing_times_df_day = pd.DataFrame({"day": missing_times.day})
                    missing_times_df_hour = pd.DataFrame({"hour": missing_times.hour})
                    missing_times_df_minute = pd.DataFrame({"minute": missing_times.minute})

                    missing_time_df = pd.concat(
                        [
                            missing_times_df_year["year"],
                            missing_times_df_month["month"],
                            missing_times_df_day["day"],
                            missing_times_df_hour["hour"],
                            missing_times_df_minute["minute"],
                        ],
                        axis=1,
                    )

                    print("Missing timeseries:")
                    print(missing_times)

                else:
                    raise KeyError(f"Note: {24 * 365 - len(time_df)} entries for an entire year are missing. Too much!")

            temperature_air_list_k = values["t2m"].tolist()
            pressure_list_pa = values["sp"].tolist()
            wind_speed_u_list_m_s = values["u10"].tolist()
            wind_speed_v_list_m_s = values["v10"].tolist()
            global_horizontal_irradiation_list_j_per_m2 = values["ssrd"].tolist()
            direct_horizontal_irradiation_list_j_per_m2 = values["fdir"].tolist()

            temperature_air_list_celsius = []
            for temperature_in_k in temperature_air_list_k:
                temperature_in_celsius = temperature_in_k - 273.15
                temperature_air_list_celsius.append(temperature_in_celsius)
            temperature_air_df = pd.DataFrame(temperature_air_list_celsius, columns=["temperature"])
            temperature_air_df = temperature_air_df.interpolate(method="linear", limit_direction="backward")

            pressure_list_hpa = []
            for pressure_in_pa in pressure_list_pa:
                pressure_in_hpa = pressure_in_pa / 100
                pressure_list_hpa.append(pressure_in_hpa)
            pressure_df = pd.DataFrame(pressure_list_hpa, columns=["pressure"])
            pressure_df = pressure_df.interpolate(method="linear", limit_direction="backward")

            wind_speed_list_m_s = []
            wind_direction_list = []
            for u_i, v_i in zip(wind_speed_u_list_m_s, wind_speed_v_list_m_s):
                wind_speed_res = np.sqrt(np.square(u_i) + np.square(v_i))
                wind_speed_list_m_s.append(wind_speed_res)

                wind_direction = 180 + (180 / np.pi) * np.arctan2(v_i, u_i)
                wind_direction_list.append(wind_direction)
            wind_speed_df = pd.DataFrame(wind_speed_list_m_s, columns=["wind_speed"])
            wind_speed_df = wind_speed_df.interpolate(method="linear", limit_direction="backward")
            wind_direction_df = pd.DataFrame(wind_direction_list, columns=["wind_direction"])
            wind_direction_df = wind_direction_df.interpolate(method="linear", limit_direction="backward")

            global_horizontal_irradiation_list_watt_per_m2 = []
            for global_horizontal_irradiation_in_j_per_m2 in global_horizontal_irradiation_list_j_per_m2:
                global_horizontal_irradiation_in_watt_per_m2 = global_horizontal_irradiation_in_j_per_m2 / (3600)
                global_horizontal_irradiation_list_watt_per_m2.append(global_horizontal_irradiation_in_watt_per_m2)
            global_irradiance_df = pd.DataFrame(
                global_horizontal_irradiation_list_watt_per_m2,
                columns=["global_irradiance"],
            )
            global_irradiance_df = global_irradiance_df.interpolate(method="linear", limit_direction="backward")

            direct_horizontal_irradiation_list_watt_per_m2 = []
            for direct_horizontal_irradiation_in_j_per_m2 in direct_horizontal_irradiation_list_j_per_m2:
                direct_horizontal_irradiation_in_watt_per_m2 = direct_horizontal_irradiation_in_j_per_m2 / (3600)
                direct_horizontal_irradiation_list_watt_per_m2.append(direct_horizontal_irradiation_in_watt_per_m2)
            direct_irradiance_df = pd.DataFrame(
                direct_horizontal_irradiation_list_watt_per_m2,
                columns=["direct_irradiance"],
            )
            direct_irradiance_df = direct_irradiance_df.interpolate(method="linear", limit_direction="backward")

            weather_df = pd.DataFrame()
            weather_df = pd.concat(
                [
                    time_df,
                    temperature_air_df,  # ["temperature"],
                    pressure_df,  # ["pressure"],
                    wind_direction_df,  # ["wind_direction"],
                    wind_speed_df,  # ["wind_speed"],
                    direct_irradiance_df,  # ["direct_irradiance"],
                    global_irradiance_df,  # ["global_irradiance"],
                ],
                axis=1,
                join="outer",
            )

            if len(missing_time_df) != 0:
                weather_df_time_index = pd.to_datetime(weather_df[["year", "month", "day", "hour", "minute"]])
                time_index_full = pd.to_datetime(missing_time_df[["year", "month", "day", "hour", "minute"]])

                merged_weather_df = pd.concat(
                    [
                        missing_time_df.set_index(time_index_full),
                        weather_df.set_index(weather_df_time_index),
                    ]
                ).sort_index()

                interpolated_weather_df = merged_weather_df.interpolate()

                weather_df = interpolated_weather_df
                print("Missing times successfully added and weather data interpolated!")

            weather_df = weather_df.reset_index(drop=True)

            # if weather_df.isna().any(axis=None):
            #     raise KeyError("The DataFrame weather_df contains NaN values. Check data")
            # if weather_df.eq("nan").any().any():
            #     raise KeyError("The DataFrame weather_df contains NaN values. Check data")

            self.safe_as_csv(weather_df, csv_path)

        self.csv_path = csv_path
